/****************************************************************************************
    Copyright 2004 uja 
    Copyright 2004 Trolltech (Qt3)
    published under GPL
    This file is part of Fleur: game implementation
*****************************************************************************************/
#include <stdio.h> // for debugging purpose
#include "chips.h"
#include <qapplication.h>

chips::chips(QWidget* parent)
{ 
  zeit=ZEIT0;
  schwerkraft=false;
  modus=0;
  for (int i=0; i<SMAX; i++) { f[i]=0; h_liste[i]=0; }				
  my_timer  =new QTimer(this); // Sekundenzaehler
  my_stopper=new QTimer(this); // fuer singleShot
  tic=200;
  connect(my_timer,SIGNAL(timeout()),this,SLOT(ticken()));                      // Anzeige alle s
  connect(this,SIGNAL(punkte_changed(int)),parent,SLOT(set_punkte(int)));       // Anzeige Punkte
  connect(this,SIGNAL(zeit_changed(int)),parent,SLOT(set_zeit(int)));           // Anzeige Restzeit
  connect(this,SIGNAL(level_changed(QString)),parent,SLOT(set_level(QString))); // Levelanzeige
  my_timer->start(1000);
}

// Listenroutinen-Auszug: (c)uja 1989
int  chips::in_liste(int e,int l[])  { int i,k=-1; if (l[0]>0) for (i=1; i<=l[0]; i++) if (e==l[i]) k=i; return k; }
void chips::add_liste(int e,int l[]) { if (in_liste(e,l)<1) { l[0]++; l[l[0]]=e; } }
void chips::merge_liste(int* l,int l1[]) { if (l1[0]>0) for (int i=1; i<=l1[0]; i++) add_liste(l1[i],l); }

// Steuerung1:
void chips::pause()       { aktiv=false; pausiert=true;  p_pause->show(); }
void chips::weiter()      { aktiv=true;  pausiert=false; p_pause->hide(); }
bool chips::is_ende()     { bool ok=true; for (int i=0; i<SMAX; i++) if (f[i]>0) ok=false; return ok; }
void chips::game_over()   { aktiv=false; p_gover->show(); emit gameover(punkte); }
void chips::reset_timer() {  zeit=ZEIT0; emit zeit_changed(zeit); }
void chips::set_cursor(int i) { p_cursor->move((i%XMAX)*DX,(i-i%XMAX)/XMAX*DY); p_cursor->show(); }


void chips::ticken()
{ 
  if (aktiv)
  { zeit--;
    emit zeit_changed(zeit);
    punkte=(zeit*multi/2);
    emit punkte_changed(punkte);
    if (zeit<0) { punkte=0; game_over(); }
  }
}

// Feld- und Pfadanzeigen:
void chips::zeige_feld(int nr) 
{ int k=f[nr];
  bool rand=false; 
  if (k==0) // Randbereich checken 
  { 
    int xx=nr%XMAX;
    rand=((xx==0) || (xx==(XMAX-1)) || (nr<XMAX) || (nr>(SMAX-XMAX)));
    if (rand) k=37;
  }
  p_sprite[nr]->setFrame(k);
}

// Auswertekette, die timergesteuert nacheinander ablaeuft:
// pfad-zeichnen, pfad_loeschen, droppen(n1)/droppen(n2)
// ----------------------------------------------------------
void chips::pfad_zeichnen() // erhaelt gefuellte Pfadliste von checkzug(int,int), Start der Auswertekette
{ 
  if (h_liste[0]>0)
  for (int i=1; i<=h_liste[0]; i++) if (f[h_liste[i]]==0) p_sprite[h_liste[i]]->setFrame(38); else f[h_liste[i]]=0;
  my_stopper->singleShot(tic,this,SLOT(pfad_loeschen()));
} 

void chips::pfad_loeschen()
{   
  for (int i=1; i<=h_liste[0]; i++) zeige_feld(h_liste[i]);
  if (is_ende()) game_over();
  else if (schwerkraft) my_stopper->singleShot(tic,this,SLOT(droppen())); 
  else aktiv=true;
}

// Modus Schwerkraft, Steinluecken beseitigen: durchfallen auf Linie (ymax-2)
// in Spalte nr von unten her besetzte Felder in genullten Vektor einsammeln,
// diesen dann in der Spalte abbilden:
void chips::droppen()
{ 
  int i,j,k,pvek,vek[YMAX];
  for (j=0; j<XMAX; j++)
  { for (i=0; i<YMAX; i++) vek[i]=0;
    pvek=YMAX-2;  // Reserveboden
    for (i=YMAX-2; i>=0; i--) { k=i*XMAX+j; if (f[k]>0) { vek[pvek]=f[k]; pvek--; } }
    for (i=0; i<YMAX; i++) { k=j+i*XMAX; f[k]=vek[i]; zeige_feld(k); }
  }
  aktiv=true;
}  
      
// Eingabecheck und Pfadliste erstellen: ------------------------------------------------------

// lineare Verbindung:
void chips::linear(int start,int ziel,int liste1[])
{
  int i,k,x1,x2,y1,y2,n1=start,n2=ziel;
  x1=n1%XMAX; y1=(n1-x1)/XMAX; x2=n2%XMAX; y2=(n2-x2)/XMAX; 
  bool ok=false;
  if (y1==y2)
  { ok=true;
    liste1[0]=0; 
    add_liste(n1,liste1);
    add_liste(n2,liste1);
    if (n2<n1) { k=n1; n1=n2; n2=k; }
    if (n2-n1>1) for (i=n1+1; i<n2; i++) if (f[i]==0) add_liste(i,liste1); else ok=false;
  }
  else if (x1==x2)
  { ok=true;
    liste1[0]=0;
    add_liste(n1,liste1);
    add_liste(n2,liste1);
    if (y2<y1) { k=y2; y2=y1; y1=k; }
    if (y2-y1>1) for (i=y1+1; i<y2; i++) { k=i*XMAX+x1; if (f[k]==0) add_liste(k,liste1); else ok=false; }
  }    
  if (!ok) liste1[0]=0;
}

// eine_ecke besteht aus 2 linearen Verbindungen. Dabei fuehrt ein moeglicher Weg ueber den Stuetzpunkt x1/y2,
// der andere ueben x2/y1. Ist ein- und derselbe Stuetzpunkt fuer beide Endpunkte zu erreichen, ist eine_ecke true. 
// arbeitet auf liste1, und liste2, merged in liste1:
void chips::eine_ecke(int n1,int n2,int liste1[])
{
  bool ok=false;
  int na,nb,x1,x2,y1,y2,xa,xb,ya,yb;
  int liste2[2*XMAX];
  x1=n1%XMAX; y1=(n1-x1)/XMAX; x2=n2%XMAX; y2=(n2-x2)/XMAX;
  xa=x1; ya=y2; xb=x2; yb=y1; na=xa+ya*XMAX; nb=xb+yb*XMAX; 
  if (f[na]==0)
  {
    linear(n1,na,liste1);
    linear(n2,na,liste2);
    ok=((liste1[0]>0) && (liste2[0]>0));
  }
  if (!ok) if (f[nb]==0)           
  { 
    linear(n1,nb,liste1);
    linear(n2,nb,liste2);
    ok=((liste1[0]>0) && (liste2[0]>0));
  }
  if (ok) merge_liste(liste1,liste2); else liste1[0]=0;
}

// ausgehend von n1 aus werden die naechsten freien Felder abgesucht, ob mit eine_ecke das Ziel erreicht werden kann:
void chips::zwei_ecken(int n1,int n2,int liste[])
{
  int na,x1,x2,y1,y2,stuetz=-1;
  int liste2[3*XMAX];
  x1=n1%XMAX; y1=(n1-x1)/XMAX; x2=n2%XMAX; y2=(n2-x2)/XMAX; liste[0]=0; liste2[0]=0;
  // Suche nach Norden:
  na=n1; while ((stuetz<0) && (na>=XMAX))
  { na=na-XMAX; 
    if (f[na]==0) { eine_ecke(na,n2,liste); if (liste[0]>0) stuetz=na; } else na=0;
  }
  // Suche nach Sueden:
  na=n1; while ((stuetz<0) && (na<SMAX-XMAX))
  { na=na+XMAX; 
    if (f[na]==0) { eine_ecke(na,n2,liste); if (liste[0]>0) stuetz=na; } else na=SMAX;
  }
  // Suche nach Osten:
  na=n1; while ((stuetz<0) && (na%XMAX<XMAX-1))
  { na++; 
    if (f[na]==0) { eine_ecke(na,n2,liste); if (liste[0]>0) stuetz=na; } else na=XMAX-1;    
  }                     
 // Suche nach Westen:
  na=n1; while ((stuetz<0) && (na%XMAX>0))
  { na--; 
    if (f[na]==0) { eine_ecke(na,n2,liste); if (liste[0]>0) stuetz=na; } else na=0;
  }
  if (stuetz>=0) 
  { linear(n1,stuetz,liste2);
    merge_liste(liste,liste2);
    add_liste(n1,liste);
  }     
}

// Check, ob Zug gueltig ist und Sammeln der zu loeschenden Steine:
bool chips::check_zug(int n1,int n2)
{ 
  h_liste[0]=0;
  if (f[n1]==f[n2]) 
  {
    linear(n1,n2,h_liste);
    if (h_liste[0]<1) eine_ecke(n1,n2,h_liste);
    if (h_liste[0]<1) zwei_ecken(n1,n2,h_liste);
  }    
  return (h_liste[0]>0);
}

// -------------------------------------------------------------------------------------------------------------------
// --- Interaktion mit User abgesehen von dem Kleinvieh ganz oben: ----------------------------------------------------

void chips::neues_spiel()
{ 
  int i,j,k,x0,y0,xlen,ylen,bmax;
  int talon[SMAX];
  k=0; for (i=0; i<SMAX; i++) { if (i%4==0) k++; talon[i]=k; }
  for (i=0; i<SMAX; i++) f[i]=0;
  h_liste[0]=0;
  srand(time(0)); 
  gewaehlt=-1;
  punkte=0;
  my_stopper->stop();
  emit punkte_changed(punkte);  

  if (modus==1)      { xlen=12; x0=3; ylen=9; y0=1; multi=3; }  
  else if (modus==2) { xlen=16; x0=1; ylen=8; y0=2; multi=3; }
  else if (modus==3) { xlen=16; x0=1; ylen=9; y0=1; multi=4; }
  else { xlen=12; x0=3; ylen=8; y0=2; multi=2; } 
  bmax=xlen*ylen;
 
  for (i=0; i<bmax; i++) { k=rand()%bmax; j=talon[i]; talon[i]=talon[k]; talon[k]=j; }     // mischen  
  k=0; for (i=0; i<xlen; i++) for (j=0; j<ylen;j++) { f[i+x0+(j+y0)*XMAX]=talon[k]; k++; } // verteilen
    
  wertung=true;
  for (i=0; i<SMAX; i++) { p_sprite[i]->show();  zeige_feld(i); }
  p_gover->hide();
  p_cursor->hide();
  reset_timer();
  weiter();
}


void chips::klick(int xx,int yy)
{ 
  if (aktiv) 
  { int x=(xx-OFX)/DX,y=(yy-OFY)/DY,nr=x+y*XMAX;
    if (f[nr]>0)
    { 
      if (gewaehlt<0) { gewaehlt=nr; set_cursor(gewaehlt); }
      else if (gewaehlt!=nr)
      { 
        p_cursor->hide();
        int n1=gewaehlt;
        gewaehlt=-1;
        aktiv=false;
        if (check_zug(n1,nr)) pfad_zeichnen(); else aktiv=true;
      }
    }
  }    
  else if (pausiert) weiter();
}

// ===== INI- und Menu2-Auswertung

void chips::set_modus(int nr)  { modus=nr; neues_spiel(); }
void chips::set_gravi(bool on) { schwerkraft=on; if (schwerkraft) emit level_changed(_ON); else emit level_changed(_OFF); }


void chips::save_settings(int hsc)
{
  QFile file(INIFILE); 
  QString buff,bu,b;
  
  buff.setNum(modus);
  if (schwerkraft) bu="1"; else bu="0";
  b.setNum(hsc);
  buff=b+"; hscpathno.\n"+buff+';'+bu+";  board mode, gravitation\n";
  if (file.open(IO_WriteOnly)) { file.writeBlock(buff,buff.length()); file.close(); }
}
// ==================================================================================================================================
